

local o = TradeskillDB_Links2; if (o.SHOULD_LOAD == nil) then return; end


local l_t_Init; -- ()
local l_t_GetLocals; -- activeDB, activeDBCount, recentlyAdded = ()
local l_t_UnregisterForAllEvents; -- ()

local l_OnEvent_TradeskillLinksMonitor1_LINKS_FOUND; -- (linksTable, source)

local l_GetActiveDB; -- db, dbCount = ()
local l_GetErrata; -- errata = ()
local l_GetRecentlyAdded; -- recent = ()

local l_IsBanned; -- banned = (ID)
local l_SetBanned; -- banned = (ID, enabled)

local l_BuildChatlink; -- chatlink = (ID[, name])
local l_CreateSpecificTypeMatchString; -- matchString = (tsPrefix, text)

local l_GetData; -- data = (ID)
local l_SetData; -- oldData, newData = (ID[, newData])

local l_LoadStaticDB; -- (staticDB)


local tonumber = tonumber; -- BuildChatlink, GetData, SetData, IsBanned, SetBanned


local l_activeDB;
local l_activeDBCount;
local l_recentlyAdded;
local l_errata;
local l_banList;


do
	if (o.t_GetLocals ~= nil) then
		l_activeDB, l_activeDBCount, l_recentlyAdded = o.t_GetLocals();
	else
		l_activeDB = {};
		l_activeDBCount = 0;
		l_recentlyAdded = {};
	end
end




function l_t_Init()
	l_t_Init = nil;
	o.t_Init = nil;
	
	
	local saved = TradeskillDB_Links2_SavedData;
	if (saved == nil) then
		saved = {};
		TradeskillDB_Links2_SavedData = saved;
	end
	
	local errata = saved.errata;
	if (errata == nil) then
		errata = {};
		saved.errata = errata;
	end
	l_errata = errata;
	
	local banList = saved.banList;
	if (banList == nil) then
		banList = {};
		saved.banList = banList;
	end
	l_banList = banList;
	
	
	local em = o.EventsManager1;
	if (o.t_UnregisterForAllEvents ~= nil) then
		o.t_UnregisterForAllEvents();
	end
	o.t_UnregisterForAllEvents = l_t_UnregisterForAllEvents;
	
	em.RegisterForEvent(l_OnEvent_TradeskillLinksMonitor1_LINKS_FOUND, nil, "TradeskillLinksMonitor1_LINKS_FOUND", false, false);
	
	local ACE = em.AddCustomEvent;
	ACE("TradeskillDB_Links2_DB_LOADED", true);
	ACE("TradeskillDB_Links2_DB_SIZE_CHANGED", false);
	ACE("TradeskillDB_Links2_ENTRY_NAME_CHANGED", false);
	ACE("TradeskillDB_Links2_ENTRY_BAN_CHANGED", false);
	
	if (o.STATIC_DB ~= nil) then
		l_LoadStaticDB(o.STATIC_DB);
		o.STATIC_DB = nil;
	end
end

o.t_Init = l_t_Init;



function l_t_GetLocals()
	l_t_GetLocals = nil;
	o.t_GetLocals = nil;
	
	return l_activeDB, l_activeDBCount, l_recentlyAdded;
end

o.t_GetLocals = l_t_GetLocals;



function l_t_UnregisterForAllEvents()
	l_t_UnregisterForAllEvents = nil;
	o.t_UnregisterForAllEvents = nil;
	
	o.EventsManager1.UnregisterForEvent(l_OnEvent_TradeskillLinksMonitor1_LINKS_FOUND, "TradeskillLinksMonitor1_LINKS_FOUND");
end




do
	local pairs = pairs;
	local stringmatch = string.match;
	
	function l_OnEvent_TradeskillLinksMonitor1_LINKS_FOUND(linksTable)
		local ID, name;
		for index, link in pairs(linksTable) do
			ID, name = stringmatch(link, "(%d+)|h%[(.+)%]", 21);
			l_SetData(ID, name);
		end
	end
end




function l_GetActiveDB()
	return l_activeDB, l_activeDBCount;
end

o.GetActiveDB = l_GetActiveDB;



function l_GetErrata()
	return l_errata;
end

o.GetErrata = l_GetErrata;



function l_GetRecentlyAdded()
	return l_recentlyAdded;
end

o.GetRecentlyAdded = l_GetRecentlyAdded;




function l_IsBanned(ID)
	return ((l_banList[tonumber(ID)] and true) or false);
end

o.IsBanned = l_IsBanned;



function l_SetBanned(ID, enabled)
	baseID = tonumber(ID);
	enabled = ((enabled and true) or nil);
	if (l_banList[ID] ~= enabled) then
		if (enabled == false) then
			l_SetData(ID, nil);
		end
		
		l_banList[ID] = enabled;
		o.EventsManager1.DispatchEvent("TradeskillDB_Links2_ENTRY_BAN_CHANGED", ID, (enabled or false));
	end
	return l_IsBanned(ID);
end

o.SetBanned = l_SetBanned;




function l_BuildChatlink(ID, name)
	ID = tonumber(ID);
	local link;
	
	if (name == nil) then
		name = l_activeDB[ID];
	end
	if (ID ~= nil and name ~= nil) then
		link = ("|cffffd000|Henchant:%d|h[%s]|h|r"):format(ID, name);
	end
	
	return link;
end

o.BuildChatlink = l_BuildChatlink;



function l_CreateSpecificTypeMatchString(tsPrefix, text)
	if (text ~= nil and text:sub(1, 1) == "^") then
		return ("^" .. tsPrefix .. text:sub(2, -1));
	else
		return ("^" .. tsPrefix .. ".*" .. (text or ".+"));
	end
end

o.CreateSpecificTypeMatchString = l_CreateSpecificTypeMatchString;




function l_GetData(ID)
	return l_activeDB[tonumber(ID)];
end

o.GetData = l_GetData;



do
	local tableinsert = table.insert;
	
	function l_SetData(ID, newData)
		ID = tonumber(ID);
		if (ID == nil) then
			error("ID is nil");
		end
		
		local oldData = l_activeDB[ID];
		
		if (l_banList[ID] == true) then
			return oldData, oldData;
		end
		
		if (oldData ~= newData) then
			l_activeDB[ID] = newData;
			l_errata[ID] = newData;
			
			if (oldData == nil and newData ~= nil) then
				local oldSize = l_activeDBCount;
				local newSize = (oldSize + 1);
				l_activeDBCount = newSize;
				tableinsert(l_recentlyAdded, 1, l_BuildChatlink(ID, newData));
				l_recentlyAdded[21] = nil;
				o.EventsManager1.DispatchEvent("TradeskillDB_Links2_DB_SIZE_CHANGED", oldSize, newSize);
			elseif (oldData ~= nil and newData == nil) then
				local oldSize = l_activeDBCount;
				local newSize = (oldSize - 1);
				l_activeDBCount = newSize;
				o.EventsManager1.DispatchEvent("TradeskillDB_Links2_DB_SIZE_CHANGED", oldSize, newSize);
			end
			
			o.EventsManager1.DispatchEvent("TradeskillDB_Links2_ENTRY_NAME_CHANGED", ID, oldData, newData);
			return oldData, newData;
		else
			return oldData, oldData;
		end
	end
	
	o.SetData = l_SetData;
end




function l_LoadStaticDB(staticDB)
	-- If this is the first static DB being loaded, we want to clean the errata and set the data directly.
	-- Otherwise, we should just use SetData as normal to add links.
	
	if (next(l_activeDB) == nil) then
		for ID, name in pairs(l_errata) do
			if (staticDB[ID] == name) then
				l_errata[ID] = nil;
			else
				staticDB[ID] = name;
			end
		end
		
		for ID in pairs(l_banList) do
			staticDB[ID] = nil;
		end
		
		local count = 0;
		for ID, name in pairs(staticDB) do
			l_activeDB[ID] = name;
			count = (count + 1);
		end
		l_activeDBCount = count;
		o.EventsManager1.DispatchEvent("TradeskillDB_Links2_DB_LOADED");
		o.EventsManager1.DispatchEvent("TradeskillDB_Links2_DB_SIZE_CHANGED", 0, count);
		
	else
		for ID, name in pairs(staticDB) do
			l_SetData(ID, name);
		end
	end
end

o.LoadStaticDB = l_LoadStaticDB;

